/**
 * @file
 * @brief The request processing, handling formatting and request modification.
 *
 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark of Netsweeper Inc.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#include <syslog.h>
#include <pthread.h>

#include "redir.h"
#include "protocol.h"
#include "config.h"
#include "request.h"
#include "utils.h"
#include "log.h"

/** @brief Flag used to signal when we should stop processing and shutdown */
volatile int redir_should_exit = 0;

/** @brief Lock used when writting a response back to Squid */
pthread_mutex_t squid_write_lock = PTHREAD_MUTEX_INITIALIZER;

/**
 * @brief allocate size of memory and terminate the entire application on error
 *
 * @param size The size of memory we want to allocate
 * @return void* Always the new memory pointer, never fails
 */
void * xmalloc(size_t size)
{
	void *value = malloc(size);
	if(value == 0) {
		log_send(LOG_ERR, "Failed to allocate memory.\n");
		exit(1);
	}
	return value;
}

/**
 * @brief duplicate a string
 *
 * @param str string to duplicate
 * @return char* pointer to the new allocated string
 */
char * xstrdup(char *str)
{
	if(str == NULL) {
		return NULL;
	}
	char *newstr = strdup(str);
	if(newstr == NULL) {
		log_send(LOG_ERR, "Failed to allocate memory.\n");
		exit(1);
	}
	return newstr;
}

/**
 * @brief free a dynamic Request request
 *
 * @param request The request we wish to free
 */
void redir_free_request(Request *request)
{
	if(request == NULL) {
		return;
	}

	if(request->id) {
		free(request->id);
	}

	if(request->URL) {
		free(request->URL);
	}

	if(request->message) {
		free(request->message);
	}

	free(request);
}

/**
 * @brief nitialize a dynamic Request request
 *
 * @return Request * The newly allocated and initialized request, dynamically allocated
 */
void * redir_init_request()
{
	Request *request = xmalloc(sizeof(Request));

	request->id = NULL;
	request->URL = NULL;
	request->message = NULL;
	request->ishttp = 1;

	request->next = NULL;
	return request;
}

/**
 * @brief Watch file descriptor with select function
 *
 * @param type, 0 or 1 for input or output filedescriptor respectively
 * @@param conf global configuration including timeout, retry, and file descriptor message
 *
 * @return int 1, 0, or -1 based on the following,
 * 1 the file descriptor is read to read/write
 * 0 the file descriptor is timeout for max_times_num
 * -1 the file descriptor is bad seriouslym should exit program
 */
static int redir_watch_fd(int type, GlobalConf *conf)
{
	int fd, sresult = 1, retry;
	fd_set fdset;
	struct timeval tv;

	retry = 0;
	while(retry++ < conf->max_timeouts_num) {
		if(type == 0) { // read file descriptor
			fd = conf->input_fd;
			tv.tv_sec = conf->squid_read_timeout / 1000;
			tv.tv_usec = (conf->squid_read_timeout % 1000) * 1000;
		} else {
			fd = conf->output_fd;
			tv.tv_sec = conf->squid_write_timeout / 1000;
			tv.tv_usec = (conf->squid_write_timeout % 1000) * 1000;
		}
		FD_ZERO(&fdset);
		FD_SET(fd, &fdset);

		if(type == 0) { // read file descriptor
			sresult = select(fd + 1, &fdset, NULL, NULL, &tv);
		} else {
			sresult = select(fd + 1, NULL, &fdset, NULL, &tv);
		}

		if(sresult == 0 || errno == EINTR) {
			sresult = 0;
		} else {
			break;
		}
	}
	return sresult;
}

/*
 * @brief Read a line from non-blocking input stream
 *
 * This function reads input char by char until newline.
 * If the buffer is overflown it skips the rest of input until newline
 * but returns an empty string.
 * If there is nothing to read from the input it waits until some chars
 * become avalibale or error happens. Each 'read timeout' it interrapts
 * to check 'should_exit' flag.
 * If some errors happen (or 'should exit' is set) returns -1;
 *
 *
 * @param fd Input file discriptor
 * @param buffer The input buffer
 * @param size The buffer free space size
 * @param timeout The timeout to check 'should exit' flag
 * @return The number of char read or -1 on error
 */

int redir_read_line(int fd, char *buffer, int size, int timeout)
{
	int ret;
	int rest = size;
	char *ptr = buffer;

	int overflow = 0;

	// we read chars from stdin until newline char or until exit flag is set
	while (1) {

		// try to read one char
		ret = read(fd, ptr, 1);

		// if we got a char
		if (ret == 1) {
			// if we got the terminating newline char
			if (*ptr == '\n') {
				if (!overflow) {
					ptr++;
				} // if buffer was overflown
				  // put zero byte right to the buffer begin
				*ptr = '\0';
				break;
			} else {
				// got a current char
				// if the buffer is not overflown
				if (! overflow) {
					ptr++;
					rest--;
					if (rest == 0) {
						overflow = 1;
						ptr = buffer;
					}
				} // otherwise we don't shift pointer
				  // and change counter, just wait newline sign
			}
		// we got "end-of-file"
		} else if (ret == 0) {
			// set th exit flag and return
			redir_should_exit = 1;
			return -1;
		// we got an error
		} else {
			// we have nothing to read on non-blocking fd
			if (errno == EAGAIN) {
				fd_set rd;
				struct timeval tv;

				// wait until get something to read or until exit flag is set
				while (1) {
					// prepare data for select
					FD_ZERO(&rd);
					FD_SET(fd, &rd);

					tv.tv_sec = timeout / 1000;
					tv.tv_usec = (timeout % 1000) * 1000;

					ret = select(fd + 1, &rd, NULL, NULL, &tv);

					// we got something to read
					if (ret == 1) {
						break;
					// timeout expired
					} else if (ret == 0) {
						// if 'should_exit' fla is set return
						if (redir_should_exit) {
							return -1;
						} // otherwise continue waiting
					// select error
					} else {
						log_send(LOG_ERR, "Select error on stdin read");
						redir_should_exit = 1;
						return -1;
					}
				}
			// something realy bad happend
			} else if (errno != EINTR) {
				log_send(LOG_ERR, "Stdin read error");
				redir_should_exit = 1;
				return -1;
			} // else the error is EINTR: do nothing just return to reading
		}
	}
	if (overflow) {
		return 0;
	}

	return size - rest;
}

/*
 * @brief Get request from Squid
 * 		return null if reading was timeout or was interrupted
 * 		exit if select return error from file descriptor
 * squid request: [ID <SP>] URL <SP> client_ip "/" fqdn <SP> user <SP> method [<SP> kvpairs]|[<sp> urlgroup]<NL>
 *
 * @param conf the global configuration passed from command line
 * @return Request* The request on success, NULL otherwise
 */
Request * redir_get_request(GlobalConf *conf)
{
	int ret;
	Request *request = NULL;
	protocol_input_line *parsed_line = NULL;
	char request_orig[conf->max_request_size + 2];

	request = redir_init_request();
	if(request == NULL) {
		goto error;
	}

	// read a line from stdin
	ret = redir_read_line(conf->input_fd, request_orig, conf->max_request_size, conf->squid_read_timeout);

	// if read error or 'should exit'
	if (ret < 0) {
		redir_free_request(request);
		return NULL;
	}

	// if buffer overflow
	if (ret == 0) {
		request_orig[0] = '\n';
		request_orig[1] = '\0';
		goto bounce;
	}

	log_send(LOG_INFO, "Received request: %d, %s\n", ret, request_orig);

	// parse input based on input protocol
	if(protocol_parse(request_orig, &parsed_line, conf->input_protocol)) {
		goto bounce;
	}

	// mangle 'domain\user' to 'user@domain', if enabled
	if(conf->username_domain_mangling) {
		if(protocol_mangle_username(parsed_line, conf)) {
			goto bounce;
		}
	}

	if (conf->ip2domain) {
		if (strchr(parsed_line->username, '@') == NULL) {
			protocol_append_domain(conf->ip2domain, parsed_line);
		}
	}

	// replace invalid characters in the username
	protocol_replace_badchars(conf, parsed_line);

	// loweracase the username if requested
	if (conf->username_lowercase) {
		protocol_lowercase_username( parsed_line);
	}

	request->message = xmalloc(conf->max_request_size + 1);
	// encode output based on output protocol
	if(protocol_encode(conf, parsed_line, request->message, conf->max_request_size + 1, conf->output_protocol)) {
		goto bounce;
	}

	// save request id
	if(parsed_line->req_id != NULL) {
		request->id = xstrdup(parsed_line->req_id);
	}

	// save request url
	if(parsed_line->url != NULL) {
		request->URL = xstrdup(parsed_line->url);
	}

	// size of the NSP request
	request->size = strlen(request->message);

	// Time to retry request.
	request->ttl = conf->max_retry_times;

	// Everything except CONNECT is http
	if ( parsed_line->method ) {
		request->ishttp = ! strcmp("CONNECT", parsed_line->method) ? 0 : 1;
	}
	else {
		request->ishttp = 1;
	}

	// destroy the input line
	protocol_parsed_line_destroy(parsed_line);

	return request; // SUCCESS

bounce:
	request->URL = xstrdup(request_orig);
	request->URL[strlen(request->URL) - 1] = '\0';
	redir_write_reply_failed(request, conf);

	// destroy the input line
	if(parsed_line) {
		protocol_parsed_line_destroy(parsed_line);
	}

	return NULL;	// BYPASS

error:

	if(request) {
		redir_free_request(request);
	}

	// destroy the input line
	if(parsed_line) {
		protocol_parsed_line_destroy(parsed_line);
	}


	return NULL;	// ERROR
}

/*
 * @brief Write reply to squid when NSP fails to process request.
 * 	The reply is the fail-close URL if it is configured with a fail-close-url.
 * 	Otherwise, the reply is the URL in the request.
 *
 * @param request the request send from squid
 * @param conf the global configuration passed from command line
 */
void redir_write_reply_failed(Request *request, GlobalConf *conf)
{
	char *reply = NULL;
	char *url;

	if(request == NULL || request->URL == NULL || conf == NULL) {
		return;
	}

	// by default we return the requested url (host) if fail
	url = request->URL;

	// Check if fail close url (host) is specified
	// the request was not CONNECT
	// this means it was not HTTP, and we need to use the fail_close_host
	if ( request->ishttp == 0 ) {
		if ( conf->fail_close_host != NULL )  {
			url = conf->fail_close_host;
		}
	}
	// the request was http based
	// in this case we use the fail_close_url
	else {
		if ( conf->fail_close_url != NULL )  {
			url = conf->fail_close_url;
		}
	}

	// if the squid uses asynchronous mode the request has an ID
	// and the reply should have this ID
	if (request->id != NULL) {
		asprintf(&reply, "%s %s\n", request->id, url);
	} else {
		asprintf(&reply, "%s\n", url);
	}

	if (reply != NULL) {
		redir_do_write_reply(request, reply, conf);
		free(reply);
	}

	return;
}

/*
 * @brief Write reply from NSP to squid
 *
 * @param request request send from squid
 * @param reply reply from NSP. It is the request URL when it is allowed or deny page URL when denied
 * @param conf the global configuration passed from command line
 */
void redir_write_reply(Request *request, char *reply, GlobalConf *conf)
{
	// if we the request had an ID but the reply has not
	// compose a new reply with the ID
	if (reply != NULL && request->id != NULL && !conf->server_concurrency) {
		char *reply_with_id;
		if (asprintf(&reply_with_id, "%s %s", request->id, reply) < 0) {
			log_send(LOG_ERR, "Memory allocation error");
			exit(1);
		}

		redir_do_write_reply(request, reply_with_id, conf);
		free(reply_with_id);

	} else {
		redir_do_write_reply(request, reply, conf);
	}
}


/*
 * @brief Write reply to squid
 *
 * @param request request send from squid
 * @param reply composed reply. It is the request URL when it is allowed or deny page URL when denied
 * @param conf the global configuration passed from command line
 */
void redir_do_write_reply(Request *request, char *reply, GlobalConf *conf)
{
	int size;
	int ret, select_result;
	char *rep = reply;

	if(reply == NULL || conf == NULL) {
		goto done;
	}

	// validate reply
	size = strlen(rep);

	if(size >= conf->max_reply_size) {
		size = conf->max_reply_size;
		rep[size - 1] = '\n';
	} else if(size > 0 && rep[size - 1] != '\n') {
		rep[size] = '\n';
		size++;
	}

	log_send(LOG_INFO, "Sending reply: %d, %s\n", size, rep);

	pthread_mutex_lock(&squid_write_lock);

	// write reply to output stream
	while (size > 0) {
		ret = write(conf->output_fd, rep, size);
		if(ret <= 0) {
			if(errno == EAGAIN) {
				select_result = redir_watch_fd(1, conf);
				if(select_result < 0) {
					log_send(LOG_ERR, "Select on Squid output failed with errorno: %d\n", errno);
					exit(1);
				} else if (select_result == 0) {
					log_send(LOG_WARNING, "Select on Squid output timeout or interrupted.\n");
					goto done;
				}
				continue; // write it again
			} else if(errno == EINTR) {
				continue; // write it again
			} else {
				log_send(LOG_ERR, "Write on Squid output failed with errorno: %d\n", errno);
				exit(1);
			}
		}
		rep += ret;
		size -= ret;
	}
done:
	pthread_mutex_unlock(&squid_write_lock);

	if( request != NULL) {
		redir_free_request(request);
	}

	return;
}
