/**
 * @file
 * @brief This test case will make sure the performance is simply.... amazing
 *
 * We generally perform the following
 * Create and setup a test case
 * Get the request rate per second
 * Perform some math on the request rate array
 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark of Netsweeper Inc.
 */
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <pthread.h>
#include <sys/time.h>

#include "CUnit/Basic.h"

#include "jopen.h"
#include "socketserver.h"

int comp(const int * a,const int * b)
{
	if (*a==*b) {
		return 0;
	}
	if (*a < *b) {
		return -1;
	}
	return 1;
}

static void connect_callback(__attribute__ ((unused)) struct client_socket* s)
{
}
static void recv_allow_callback(struct client_socket* s)
{
	if ( strchr(s->rxbuffer, '\n' ) ) {
		// echo it back
		socketserver_send(s, s->rxbuffer, s->rxlen);
		s->rxlen=0;
	}
}

static void recv_disconnect_callback(struct client_socket* s)
{
	if ( strchr(s->rxbuffer, '\n' ) ) {
		// echo it back
		socketserver_send(s, s->rxbuffer, s->rxlen);
		s->rxlen=0;
	}

	// Shutdown this connection
	socketserver_connection_shutdown(s);
}
static void disconnect_callback(__attribute__ ((unused)) struct client_socket* s)
{
}

/**
 * @brief Our general performance testing framework function
 *
 * This function will run a performance test for a period of time
 * We make sure that factor of the mean is less then the standard deviation
 *
 * This is basically a measurement for a stable server in regards to performance.
 * This test WILL FAIL on a server or workstation with other tasks eating up resources.
 *
 * @param secs Number of seconds to run the test for
 * @param svr_callback The callback function to the sample server we create
 * @param factor The factor of the mean the stdev must be less then, 0.05 is good
 *
 */
static void performance_framework(int secs, SOCKETSERVER_CALLBACK svr_callback, double factor)
{
	struct socketserver* svr;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[1024];
	int ret;
	char* ptr;

	// Create a new amazing server
	svr = socketserver_create(12576, connect_callback, svr_callback, disconnect_callback);

	printf("%s(): Waiting for server startup\n", __FUNCTION__);

	fp = jopen("../redir localhost:12576 -f http://failcloseurl -L3");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	int* rate;
	rate = malloc(sizeof(int) * secs);
	time_t current_time;
	time_t rate_time;
	time_t start_time;
	int counter=0;
	for(rate_time=start_time=current_time=time(NULL); 1; current_time=time(NULL) ) {

		// first we want to store the request rate
		// once it is calculated
		if ( current_time != rate_time ) {
			// do not keep the first sample
			// maybe a quarter second sample
			// also allows for "warmup"
			if ( rate_time != start_time ) {
				// print
				printf("real time rate update: %d\n", counter);
				// store
				rate[(rate_time-1) - start_time] = counter;
			}

			// move
			rate_time = current_time;

			// stop the test when we moved to far
			if ( (rate_time-1) - start_time >= secs ) {
				break;
			}
			counter=0;
		}
		counter++;
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", counter);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
		ret=read(fp->from_stdout, buffer, 1023);
		if ( ret >= 0 ) {
			buffer[ret]='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(sendbuffer, ' ');
		*ptr='\0';

		// truncate sendbuffer at first space
		ptr=strchr(buffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(buffer, ' ');
		if ( ptr ) {
			*ptr='\0';
		}

		if ( strcmp(buffer, sendbuffer) ) {
			printf("ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
			CU_FAIL("We did not get the response we expected.");
			socketserver_signal_shutdown(svr);
			socketserver_waitfor_shutdown(svr);
			jclose(fp);
			return;
		}
	}


	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);
	jclose(fp);

	// now we want to calculate some stats over the array of rates
	// we have collected.  

	// sort
	qsort(rate, secs, sizeof(int), (int (*)(const void *, const void *))comp);

	int mean =0;
	for(i=0; i<secs; i++) {
		mean+=rate[i];
	}
	mean /= secs;


	// standard deviation average of the root of the sum of squares of the distance from the mean of the samples
	// with some brackets in there
	int stddev=0;
	for(i=0; i<secs; i++) {
		stddev += (rate[i] - mean) * (rate[i] - mean);
	}
	stddev /= (secs-1);
	stddev = sqrt(stddev);

	printf("mean %d stddev %d\n", mean, stddev);
	printf("%lf%% of mean is %lf\n", factor * 100.0, (double)mean * 0.05);
	
	// If 5% of the mean is less then the standard deviation we got an issue
	if ( ((double)mean * factor) < (double)stddev ) {
		printf("%lf%% of mean is < stddev\n", factor * 100.0);
		CU_FAIL("factor of the mean was less then the stardard devision");
		return;
	}

	CU_PASS("The Simple Squid Redir test has passed");
	return;
}

/**
 * @brief Our general performance testing framework function
 *
 * This function will run a performance test for a period of time
 * We make sure that factor of the mean is less then the standard deviation
 *
 * This is basically a measurement for a stable server in regards to performance.
 * This test WILL FAIL on a server or workstation with other tasks eating up resources.
 *
 * @param secs Number of seconds to run the test for
 * @param svr_callback The callback function to the sample server we create
 * @param factor The factor of the mean the stdev must be less then, 0.05 is good
 *
 */
static void performance_concurrent_framework(int secs, SOCKETSERVER_CALLBACK svr_callback, double factor)
{
	struct socketserver* svr1=NULL;
	struct socketserver* svr2=NULL;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[1025];
	int ret;
	char* ptr;
	int sent;
	int sendcount=0;
	int recvcount=0;
	int nfds;
	fd_set readers;
	fd_set writers;
	char* end;


	// Create two servers and always allow requests
	svr1 = socketserver_create(12345, connect_callback, svr_callback, disconnect_callback);

	// create the second server if we require it
	svr2 = socketserver_create(12346, connect_callback, svr_callback, disconnect_callback);
	fp = jopen("../redir localhost:12345 localhost:12346 -f http://failcloseurl -i squid2.6 -o squid2.6");

	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr1);
		socketserver_waitfor_shutdown(svr1);
		socketserver_signal_shutdown(svr2);
		socketserver_waitfor_shutdown(svr2);
		return;
	}

	int* rate;
	rate = malloc(sizeof(int) * secs);
	time_t current_time;
	time_t rate_time;
	time_t start_time;
	int counter=0;
	for(rate_time=start_time=current_time=time(NULL); 1; current_time=time(NULL) ) {
		// first we want to store the request rate
		// once it is calculated
		if ( current_time != rate_time ) {
			// do not keep the first sample
			// maybe a quarter second sample
			// also allows for "warmup"
			if ( rate_time != start_time ) {
				// print
				printf("real time rate update: %d\n", counter);
				// store
				rate[(rate_time-1) - start_time] = counter;
			}

			// move
			rate_time = current_time;

			// stop the test when we moved to far
			if ( (rate_time-1) - start_time > secs ) {
				break;
			}
			counter=0;
		}

		// we select to find out if we can read/write to the redir
		// when we can write we write (if we need to)
		// and when we can read we read... (because we always need to)
		FD_ZERO(&readers);
		FD_SET(fp->from_stdout, &readers);
		FD_ZERO(&writers);
		nfds = fp->from_stdout;
		FD_SET(fp->to_stdin, &writers);
		if ( fp->to_stdin > nfds ) {
			nfds=fp->to_stdin;
		}

		ret = select(nfds+1, &readers, &writers, NULL, 0);
		if ( ret <= 0 ) {
			socketserver_signal_shutdown(svr1);
			socketserver_waitfor_shutdown(svr1);
			socketserver_signal_shutdown(svr2);
			socketserver_waitfor_shutdown(svr2);
			jclose(fp);
			CU_FAIL("Select failure, may indicate child death of redir");
			return;
		}

		// we can send something
		if ( FD_ISSET(fp->to_stdin, &writers) )
		{
			// Chunking send
			sprintf(sendbuffer, "%d http://google.ca/%d -/- - - -\n", sendcount, sendcount);
			/*
			printf("SENDING: %s\n", sendbuffer);
			printf("--------------------------------------------------------------------------------\n");
			*/
			sent=0;
			do {
				ret = write(fp->to_stdin, sendbuffer+sent, strlen(sendbuffer) - sent);
				if ( ret > 0 ) {
					sent+=ret;
				}
				else {
					// error out
					socketserver_signal_shutdown(svr1);
					socketserver_waitfor_shutdown(svr1);
					socketserver_signal_shutdown(svr2);
					socketserver_waitfor_shutdown(svr2);
					CU_FAIL("Could not send request");
					return;
				}
			} while ( sent != (int)strlen(sendbuffer) );
			sendcount++;
		}


		// Check to see if we can recv a little
		if ( FD_ISSET(fp->from_stdout, &readers) ) {
			int len;
			// Read chunks of data and process them
			// Note: one read may read many requests, so we loop through process
			// the requests

			// Read and terminate string
			len = strlen(buffer);
			if (len >= 1024) {
				end = buffer+strlen(buffer);
				ret=read(fp->from_stdout, buffer +strlen(buffer), 1024 - strlen(buffer));
				if ( ret >= 0 ) {
					*(end+ret)='\0';
				}
				
				// each response is \n terminated
				// so we process each one
				while ( (end = strchr(buffer, '\n')) ) {
					*end = '\0';
					end++;
					
					ptr=strchr(buffer, ' ');
					*ptr='\0';
					
					// move the buffer
					recvcount++;
					memmove(buffer, end, strlen(end)+1);
					counter++;
				}
			}
			else
				break;
		}
	};

	socketserver_signal_shutdown(svr1);
	socketserver_signal_shutdown(svr2);
	socketserver_waitfor_shutdown(svr1);
	socketserver_waitfor_shutdown(svr2);
	jclose(fp);



	// now we want to calculate some stats over the array of rates
	// we have collected.  

	// sort
	qsort(rate, secs, sizeof(int), (int (*)(const void *, const void *))comp);

	int mean =0;
	for(i=0; i<secs; i++) {
		mean+=rate[i];
	}
	mean /= secs;


	// standard deviation average of the root of the sum of squares of the distance from the mean of the samples
	// with some brackets in there
	int stddev=0;
	for(i=0; i<secs; i++) {
		stddev += (rate[i] - mean) * (rate[i] - mean);
	}
	stddev /= secs -1;
	stddev = sqrt(stddev);

	printf("mean %d stddev %d\n", mean, stddev);
	printf("%lf%% of mean is %lf\n", factor * 100.0, (double)mean * 0.05);
	
	// If 5% of the mean is less then the standard deviation we got an issue
	if ( ((double)mean * factor) < (double)stddev ) {
		printf("%lf%% of mean is < stddev\n", factor * 100.0);
		CU_FAIL("factor of the mean was less then the stardard devision");
		return;
	}

	CU_PASS("The Multiple Server Redir test has passed");
	return;
}

static void performance_disconnect_test()
{
	performance_framework(30, recv_disconnect_callback, 0.05);
}

static void performance_good_test()
{
	performance_framework(30, recv_allow_callback, 0.05);
}

static void performance_2good_test()
{
	performance_concurrent_framework(30, recv_allow_callback, 0.05);
}



/**
 * @brief Performance latency test framework.
 *
 *
 * @param secs Number of seconds to run the test for
 * @param svr_callback The callback function to the sample server we create
 * @param factor The factor of the mean the stdev must be less then, 0.05 is good
 *
 */
static void performance_concurrentLatency_framework(int requests, int outstanding, SOCKETSERVER_CALLBACK svr_callback, double factor)
{
	struct socketserver* svr1=NULL;
	struct socketserver* svr2=NULL;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[1025];
	int ret;
	char* ptr;
	int sent;
	int sendcount=0;
	int recvcount=0;
	int nfds;
	fd_set readers;
	fd_set writers;
	char* end;


	// Create two servers and always allow requests
	svr1 = socketserver_create(12345, connect_callback, svr_callback, disconnect_callback);

	// create the second server if we require it
	svr2 = socketserver_create(12346, connect_callback, svr_callback, disconnect_callback);
	fp = jopen("../redir localhost:12345 localhost:12346 -f http://failcloseurl -i squid2.6 -o squid2.6");

	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr1);
		socketserver_waitfor_shutdown(svr1);
		socketserver_signal_shutdown(svr2);
		socketserver_waitfor_shutdown(svr2);
		return;
	}

	struct timeval* send_time;
	struct timeval* recv_time;
	int* latency;
	send_time = malloc(sizeof(struct timeval) * requests);
	recv_time = malloc(sizeof(struct timeval) * requests);
	latency = malloc(sizeof(int) * requests);
	do {
		// we select to find out if we can read/write to the redir
		// when we can write we write (if we need to)
		// and when we can read we read... (because we always need to)
		FD_ZERO(&readers);
		FD_SET(fp->from_stdout, &readers);
		FD_ZERO(&writers);
		nfds = fp->from_stdout;
		FD_SET(fp->to_stdin, &writers);
		if ( fp->to_stdin > nfds ) {
			nfds=fp->to_stdin;
		}

		ret = select(nfds+1, &readers, &writers, NULL, 0);
		if ( ret <= 0 ) {
			socketserver_signal_shutdown(svr1);
			socketserver_waitfor_shutdown(svr1);
			socketserver_signal_shutdown(svr2);
			socketserver_waitfor_shutdown(svr2);
			jclose(fp);
			CU_FAIL("Select failure, may indicate child death of redir");
			return;
		}

		// Check to see if we can recv a little
		if ( FD_ISSET(fp->from_stdout, &readers) ) {
			// Read chunks of data and process them
			// Note: one read may read many requests, so we loop through process
			// the requests

			// Read and terminate string
			end = buffer+strlen(buffer);
			ret=read(fp->from_stdout, buffer +strlen(buffer), 1024 - strlen(buffer));
			if ( ret >= 0 ) {
				*(end+ret)='\0';
			}

			// each response is \n terminated
			// so we process each one
			while ( (end = strchr(buffer, '\n')) ) {
				*end = '\0';
				end++;

				ptr=strchr(buffer, ' ');
				*ptr='\0';


				gettimeofday(&recv_time[atoi(buffer)], NULL);

				// move the buffer
				recvcount++;
				memmove(buffer, end, strlen(end)+1);
			}
		}


		// we can send something and we have not yet sent all requests
		if ( FD_ISSET(fp->to_stdin, &writers) && sendcount < requests && ( sendcount - recvcount ) < outstanding )
		{
			// Chunking send
			sprintf(sendbuffer, "%d http://google.ca/%d -/- - - -\n", sendcount, sendcount);
			gettimeofday(&send_time[sendcount], NULL);
			sent=0;
			do {
				ret = write(fp->to_stdin, sendbuffer+sent, strlen(sendbuffer) - sent);
				if ( ret > 0 ) {
					sent+=ret;
				}
				else {
					// error out
					socketserver_signal_shutdown(svr1);
					socketserver_waitfor_shutdown(svr1);
					socketserver_signal_shutdown(svr2);
					socketserver_waitfor_shutdown(svr2);
					CU_FAIL("Could not send request");
					return;
				}
			} while ( sent != (int)strlen(sendbuffer) );
			sendcount++;
		}


	} while (sendcount < requests || recvcount < requests);

	socketserver_signal_shutdown(svr1);
	socketserver_signal_shutdown(svr2);
	socketserver_waitfor_shutdown(svr1);
	socketserver_waitfor_shutdown(svr2);
	jclose(fp);


	// Make an array of latency
	for (i = 0; i < requests; i++){
		struct timeval diff;
		timersub(&recv_time[i], &send_time[i], &diff);
		latency[i] = (diff.tv_sec * 1000000) + ( diff.tv_usec );
	}


	// now we want to calculate some stats over the array of rates
	// we have collected.  


	// Calculate the mean (average)
	long double mean =0;
	for(i=0; i<requests; i++) {
		mean+=latency[i];
	}
	mean /= requests;


	// standard deviation average of the root of the sum of squares of the distance from the mean of the samples
	// with some brackets in there
	long double stddev=0;
	long double min=latency[0];
	long double max=latency[0];
	for(i=0; i<requests; i++) {
		if ( min > latency[i] ) {
			min = latency[i];
		}
		if ( max < latency[i] ) {
			max = latency[i];
		}
		stddev += (latency[i] - mean) * (latency[i] - mean);
	}
	stddev /= (long double)(requests -1);
	stddev = sqrtl(stddev);


	printf("\nRESULTS\n");
	printf("%Lf min\n", min);
	printf("%Lf max\n", max);
	printf("%Lf mean\n", mean);
	printf("%Lf stddev\n", stddev);

	// calculate the coefficent of of variaion
	// stddev/mean
	long double cv = stddev / mean;
	printf("%Lf Coefficent of variation\n", cv);


	// Latency Jitter Calculation
	// divide the sample range into 10 hash tables
	// make sure the middle 2-9 buckets have > cv of the data
	int buckets[10];
	int bucket;
	int brange = max - min;

	for(i=0; i<10;i++) {
		buckets[i]=0;
	}

	printf("range: %d\n", brange);
	for(i=0; i<requests;i++) {
		bucket=(((long double)(latency[i] - min )) / (long double)brange) * 10.0;
		buckets[bucket]++;
	}

	// all buckets except 0 and 9 the "outliers"
	printf("Histogram of Latency\n");
	for(i=0; i<10;i++) {
		printf("%d - %d: %d\n", 
				(int)(min + ((long double)brange / 10.0)) * i,
				(int)(min + ((long double)brange / 10.0)) * (i+1),
				buckets[i]);
	}

	// Make sure the average is in the first 2 buckets
	// the first bucket maybe "bottom" liers in some cases
	// generally, all should be in the first bucket
	CU_ASSERT(mean < (int)(min + ((long double)brange / 10.0 * 2)));

	// Make sure factor of the requests are in first bucket
	CU_ASSERT(requests * factor * 100 > buckets[0]);

	

	CU_PASS("The Multiple Server Redir test has passed");
	return;
}

/**
 * @brief Create 10,000 requests, having at most 2 requests outstanding, make sure the latency stdv dev is low
 */
void performance_2good_latency_test()
{
	// we have 2 servers with 1 connection each
	// so lets say we only queue up 2 meaning nothing is concurrent really
	performance_concurrentLatency_framework(10000, 2, recv_allow_callback, 0.9);
}

/**
 * @brief Create 10,000 requests, having at most 10 requests outstanding, this is just crazy.
 */
void performance_2good_10latency_test()
{
	// we have 2 servers with 1 connection each
	// since we have 10 outstanding the latency will look something like
	// 1
	// 1
	// 2
	// 2
	// 3
	// 3
	// 4
	// 4
	// 5
	// 5
	// .. note, here we start processing entries added when we get 1 back
	// .. so really we should have stabilized at 5s with a some minor
	// ...variations... since get will process bulks of data
	// 5
	// 5
	// 5
	// 5
	// 5
	// 5
	//
	performance_concurrentLatency_framework(10000, 10, recv_allow_callback, 0.9);
}

int performance_test_init()
{
	CU_pSuite pSuite = NULL;

	// This is our test suite for test4
	pSuite = CU_add_suite("Performance Test Suite", NULL, NULL);
	if (NULL == pSuite) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	if ( (NULL == CU_add_test(pSuite, "Disconnecting Server Performance", performance_disconnect_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	if ( (NULL == CU_add_test(pSuite, "Good Server Performance", performance_good_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	if ( (NULL == CU_add_test(pSuite, "2 Good Servers Performance", performance_2good_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	if ( (NULL == CU_add_test(pSuite, "2 Servers 2 requests at a time latency", performance_2good_latency_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	if ( (NULL == CU_add_test(pSuite, "2 Servers 10 requests at a time latency", performance_2good_10latency_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}


	return 0;
}
